#ifndef SLAM_H
#define SLAM_H

#include <mrpt/slam/CICP.h>
#include <iostream>
#include <float.h>
#include <sstream>
#include <string>
#include <cmath>
#include <fstream>
#include <stdlib.h>
#include <algorithm>


#include <mrpt/poses/CPose3DPDF.h>
#include <mrpt/gui/CDisplayWindow3D.h>

#include <mrpt/maps/CColouredPointsMap.h>
#include <mrpt/opengl/CPointCloudColoured.h>
#include <mrpt/opengl/CGridPlaneXY.h>

#include <mrpt/obs/CObservation3DRangeScan.h>
#include <mrpt/math/CMatrix.h>

using namespace mrpt::poses;
using namespace mrpt::math;
using namespace mrpt::gui;
using namespace mrpt::utils;
using namespace mrpt::obs;
using namespace mrpt::maps;
using namespace std;
using namespace mrpt::slam;
using mrpt::format;

///Scan vertical resolution
#define SCAN_X 376
///Scan horizontal resolution
#define SCAN_Y 672


///Struct holding timestamped data
struct data_entry {
    size_t timestamp;
    CPose3D data;
};

class Slam {
public:
    Slam();

    Slam(const Slam &orig);

    virtual ~Slam();

    ///Starts the 3D SLAM algorithm.
    ///@param path to dataset
    ///@return vector containing transformations between each frame.
    void performSlam(string path, bool use_odometry);

private:
    ///Camera parameters.

    ///Global pose, every new transformation between two scans is appended to this pose.
    CPose3D global;
    ///Map storing the previous scan,
    CColouredPointsMap previousScan;
    ///Map storing the current scan.
    CColouredPointsMap currentScan;
    ///Stores list of computed transformations.
    vector<CPose3D> transformations;
    ///Stores list of timestamps.
    vector<size_t> timestamps;
    ///ICP algorithm class with parameters set on creation.
    CICP icp;
    ///Running time
    double RUN_TIME;

    ///Creates CObservation3DRangeScanPtr scan.
    ///@param empty CObservation3DRangeScanPtr, serves as an output.
    ///@param string used to create image address for loading.
    ///@param depth map that will be matched with 2D image to create 3D scan.
    void makeTheScan(CObservation3DRangeScanPtr scan, string imageName, float (&depth)[SCAN_X][SCAN_Y]);

    ///Computes 3D 6DOF transformation between two CColouredPointsMap.
    ///@param reference CColouredPointsMap aligned to beginning of the coordinate system.
    ///@param CColouredPointsMap to be aligned.
    ///@param serves as an output CPose3D.
    void matchScans(CColouredPointsMap &referenceScan, CColouredPointsMap &alignedScan, CPose3D &calculatedPose,
                    CPose3D &seed);

    ///Function loads depth data from binary file.
    ///@param Path to depth data.
    ///@param serves as an output for depth data.
    void loadDepthData(string depthName, float (&depth)[SCAN_X][SCAN_Y]);

    ///Function saves output of slam.
    ///@param Destination
    void save_output(string dest);

    ///Function loads timestamps from file.
    ///@param Path to timestamps data.
    void load_timestamps(string input);

    ///Function loads odometry from file.
    ///@param Path to ooometry data.
    void load_odometry(string input, std::vector<data_entry> &dataset);

    /// Finds closest matching odometry for  \param timestamp of 3D scan
    /// \param timestamp
    /// \param odometry
    /// \return odometry CPose3D
    CPose3D find_matching_odometry(size_t timestamp, std::vector<data_entry> &odometry);

};

#endif /* SLAM_H */

